import logging
import xml.etree.ElementTree as ET

from collections import defaultdict

from mwlib.xhtmlwriter import MWXHTMLWriter, SkipChildren
from mwlib import xmltreecleaner
from mwlib.advtree import Reference
xmltreecleaner.childlessOK.append(Reference)

import tex

EXCLUDE_CLASSES = frozenset(('navbox', 'collapsible', 'autocollapse',
                             'plainlinksneverexpand', 'navbar', 'metadata',
                             'navigation-box', 'stub', 'template-documentation',
                             'portal', 'NavFrame', 'NavHead', 'NavContent', 'thumbinner',
                             'thumbcaption', 'magnify', 'vertical-navbox',
                             'tmbox', 'maptable', 'printonly',
                             'fmbox', 'ombox', 'cmbox', 'sisterproject', 
                             'interProject',
                             'wikilien_alternatif', #frwiki sister projects
                             'bandeau', #frwiki message at the top of article
                             'itwiki_template_avviso', #itwiki
                             'itwiki_template_toc',
                             'itwiki_template_disclaimer_v',
                             ))

EXCLUDED_IDS = frozenset(('interProject',))

log = logging.getLogger(__name__)

#latex can render most math, generates nice transparent PNG, but fails on some formulas
#blahtex can render some formulas that latex can't, also renders transparent PNG
#texvc is what Wikipedia uses, can render the most, but output PNG is not transparent and
#doesn't looks as good as latex or blahtex
mathcmds = ('latex', 'blahtex', 'texvc')

class XHTMLWriter(MWXHTMLWriter):

    paratag = 'p'

    def __init__(self, *args, **kwargs):
        MWXHTMLWriter.__init__(self, *args, **kwargs)
        #keep reference list for each group serparate
        self.references = defaultdict(list)
        #also keep named reference positions, separate for each group
        #map named reference to 2-tuple of position first seen and count
        self.namedrefs = defaultdict(dict)

    def xwriteArticle(self, a):
        e = ET.Element("div")
        h = ET.SubElement(e, "h1")
        h.text = a.caption
        self.writeChildren(a, e)
        return SkipChildren(e)

    def xwriteChapter(self, obj):
        e = ET.Element("div")
        h = ET.SubElement(e, "h1")
        self.write(obj.caption)
        return e

    def xwriteSection(self, obj):
        e = ET.Element("div")
        level = 2 + obj.getLevel() # starting with h2
        h = ET.SubElement(e, "h%d" % level)
        self.write(obj.children[0], h)
        obj.children = obj.children[1:]
        return e

    def xwriteTimeline(self, obj):
        s = ET.Element("object")
        s.set("type", "application/mediawiki-timeline")
        s.set("src", "data:text/plain;charset=utf-8,%s" % obj.caption)
        em = ET.SubElement(s, "em")
        em.text = u"Timeline"
        return s

    def xwriteHiero(self, obj): # FIXME parser support
        s = ET.Element("object")
        s.set("type", "application/mediawiki-hiero")
        s.set("src", "data:text/plain;charset=utf-8,%s" % obj.caption)
        em = ET.SubElement(s, "em")
        em.text = u"Hiero"
        return s

    def xwriteMath(self, obj):
        for cmd in mathcmds:
            try:
                imgurl = 'data:image/png;base64,' + tex.toimg(obj.caption, cmd)
            except tex.MathRenderingFailed, e:
                log.warn('Could not render math in %r with %r: %s',
                         obj.getParents()[0].caption, cmd, e)
            except:
                log.warn('Could not render math in %r with %r',
                         obj.getParents()[0].caption, cmd, exc_info=1)
            else:
                s = ET.Element("img")
                s.set("src", imgurl)
                s.set("class", "tex")
                return s
        log.error('Failed to render math %r in %r',
                  obj.caption, obj.getParents()[0].caption)
        s = ET.Element("span")
        s.text = obj.caption
        s.set("class", "tex")
        return s

    def xwriteURL(self, obj):
        a = ET.Element("a", href=obj.caption)
        if not obj.children:
            a.text = obj.caption
        return a

    def xwriteNamedURL(self, obj):
        a = ET.Element("a", href=obj.caption)
        if not obj.children:
            name = "[%s]" % self.namedLinkCount
            self.namedLinkCount += 1
            a.text = name
        return a

    def xwriteSpecialLink(self, obj): # whats that?
        a = ET.Element("a", href=obj.url or "#")
        if not obj.children:
            a.text = obj.target
        return a

    def writeLanguageLinks(self):
        pass

    def xwriteImageLink(self, obj):
        return SkipChildren()

    def xwriteImageMap(self, obj):
        return SkipChildren()

    def xwriteGallery(self, obj):
        return SkipChildren()

    def xwriteLink(self, obj):
        a = ET.Element("a", href=obj.target)
        if not obj.children:
            a.text = obj.target
        return a

    def xwriteItemList(self, lst):
        for li in lst.children:
            if any(c.ns == 10 for c in li.children):
                return SkipChildren()
        if lst.numbered:
            tag = "ol"
        else:
            tag = "ul"
        return ET.Element(tag)

    xwriteArticleLink = xwriteLink
    xwriteInterwikiLink = xwriteLink
    xwriteNamespaceLink = xwriteLink

    def xwriteCategoryLink(self, obj):
        return SkipChildren()

    def xwriteTable(self, obj):
        tableclasses = obj.attributes.get('class', '').split()
        if any((tableclass in EXCLUDE_CLASSES for tableclass in tableclasses)):
            return SkipChildren()
        id_attr = obj.attributes.get('id', '')
        if id_attr in EXCLUDED_IDS:
            return SkipChildren()
        return MWXHTMLWriter.xwriteTable(self, obj)

    def xwriteGenericElement(self, obj):
        classes = obj.attributes.get('class', '').split()
        if any((cl in EXCLUDE_CLASSES for cl in classes)):
            return SkipChildren()
        id_attr = obj.attributes.get('id', '')
        if id_attr in EXCLUDED_IDS:
            return SkipChildren()
        return MWXHTMLWriter.xwriteGenericElement(self, obj)

    xwriteEmphasized = xwriteGenericElement
    xwriteStrong = xwriteGenericElement
    xwriteSmall = xwriteGenericElement
    xwriteBig = xwriteGenericElement
    xwriteCite = xwriteGenericElement
    xwriteSub = xwriteGenericElement
    xwriteSup = xwriteGenericElement
    xwriteCode = xwriteGenericElement
    xwriteBreakingReturn = xwriteGenericElement
    xwriteHorizontalRule = xwriteGenericElement
    xwriteTeletyped = xwriteGenericElement
    xwriteDiv = xwriteGenericElement
    xwriteSpan = xwriteGenericElement
    xwriteVar= xwriteGenericElement
    xwriteRuby = xwriteGenericElement
    xwriteRubyBase = xwriteGenericElement
    xwriteRubyParentheses = xwriteGenericElement
    xwriteRubyText = xwriteGenericElement
    xwriteDeleted = xwriteGenericElement
    xwriteInserted = xwriteGenericElement
    xwriteTableCaption = xwriteGenericElement
    xwriteDefinitionList = xwriteGenericElement
    xwriteDefinitionTerm = xwriteGenericElement
    xwriteDefinitionDescription = xwriteGenericElement
    xwriteFont = xwriteGenericElement

    def xwriteReference(self, obj):
        assert obj is not None
        group = obj.attributes.get(u'group', '')
        group_references = self.references[group]

        a = ET.Element("a")
        ref_name = obj.attributes.get('name')
        if ref_name:
            ref_name = ref_name.replace(' ', '_')
            group_namedrefs = self.namedrefs[group]
            named_ref_first, named_ref_count = group_namedrefs.get(ref_name, (None, 0))
            if named_ref_first is None:
                log.debug('adding named ref %r', ref_name)
                group_references.append(obj)
                named_ref_first = len(group_references)
            elif obj.children:
                group_references[named_ref_first-1] = obj
            note_seq_num = named_ref_first
            noteid = self.mknoteid(group, named_ref_first)
            refid = u'_r'+noteid+u'_'+str(named_ref_count)
            named_ref_count += 1
            group_namedrefs[ref_name] = (named_ref_first, named_ref_count)
        else:
            group_references.append(obj)
            note_seq_num = len(group_references)
            noteid = self.mknoteid(group, note_seq_num)
            refid = u'_r'+noteid

        a.text = u'%s %s' % (group, unicode(note_seq_num))
        a.text = u'[%s]' % a.text.strip()
        a.set('id', refid)
        a.set('href', '#')
        a.set('onClick',
              'return s(\'%s\')' % noteid)
        return SkipChildren(a)

    def xwriteReferenceList(self, t):
        if not self.references:
            return
        group = t.attributes.get('group', '')
        references = self.references.pop(group, [])
        if not references:
            return
        ol = ET.Element("ol")
        group_namedrefs = self.namedrefs.pop(group, {})
        for i, ref in enumerate(references):
            noteid = self.mknoteid(group, i+1)
            li = ET.SubElement(ol, "li", id=noteid)
            ref_name = ref.attributes.get('name')
            b = ET.SubElement(li, "b")
            b.tail = ' '

            if ref_name:
                ref_name = ref_name.replace(' ', '_')
                if not ref.children:
                    log.debug('No definition for named ref %r', ref_name)
                    #named reference has not been defined yet
                    #expect definition to be child of this reference list
                    matching_named_refs = [child for child in t.getAllChildren()
                                           if child.attributes.get('name', '').replace(' ', '_')
                                           == ref_name and child.children]
                    if matching_named_refs:
                        ref = matching_named_refs[0]
                        log.debug('Found defintion for named ref %r', ref_name)
                    else:
                        log.warn('Definition for named ref %r not found', ref_name)
                name_ref_count = group_namedrefs.get(ref_name, (None, 1))[1]
                if name_ref_count == 1:
                    ref_id = u'_r'+noteid+u'_'+unicode(0)
                    backref = ET.SubElement(b, 'a', href=u'#'+ref_id)
                    backref.set('onClick',
                                'return s(\'%s\')' % (ref_id))
                    backref.text = u'\u2191'
                else:
                    b.text = u'\u2191 '
                    backref_parent = ET.SubElement(b, 'sup')
                    for j in range(name_ref_count):
                        ref_id = u'_r'+noteid+u'_'+unicode(j)
                        backref = ET.SubElement(backref_parent, 'a', href=u'#'+ref_id)
                        backref.set('onClick',
                                    'return s(\'%s\')' % (ref_id))
                        backref.text = unicode(j+1)
                        backref.tail = ' '
            else:
                ref_id = u'_r'+noteid
                backref = ET.SubElement(b, 'a', href=u'#'+ref_id)
                backref.set('onClick',
                            'return s(\'%s\')' % (ref_id))
                backref.text = u'\u2191'
            self.writeChildren(ref, parent=li)
        return SkipChildren(ol)

    def mknoteid(self, group, num):
        return u'_n'+u'_'.join((group, unicode(num)))

    def xwriteParagraph(self, obj):
        """
        currently the parser encapsulates almost anything into paragraphs,
        but XHTML1.0 allows no block elements in paragraphs.
        therefore we use the html-div-element.

        this is a hack to let created documents pass the validation test.
        """
        e = ET.Element(self.paratag) # "div" or "p"
        return e

    def xwriteOverline(self, s):
        e = ET.Element("span")
        e.set("class", "o")
        return e

    def xwriteUnderline(self, s):
        e = ET.Element("span")
        e.set("class", "u")
        return e

    def xwriteSource(self, s):
        e = ET.Element("code")
        return e

    def xwriteCenter(self, s):
        e = ET.Element("span")
        e.set("class", "center")
        return e

    def xwriteStrike(self, s):
        e = ET.Element("del")
        return e

    def xwriteBlockquote(self, s):
        return ET.Element("blockquote")

    def xwriteIndented(self, s):
        e = ET.Element("blockquote")
        e.set("class", "indent")
        return e


def remove_childless_elements(element, parent=None):
    """
    Remove elements that are supposed to have children but are empty

    """   

    for child in element.getchildren()[:]:
        remove_childless_elements(child, element)

    if (element.tag.lower() not in ("br", "td", "img", "hr", "col", "colgroup") and
        not (element.getchildren() or element.text or element.tail) and parent):
        parent.remove(element)


def convert(obj, rtl=False):
    w = XHTMLWriter()
    e = w.write(obj)
    remove_childless_elements(e)
    if rtl:
        e.set("dir", "rtl")
    if w.languagelinks:
        languagelinks = [(obj.namespace, obj.target) for obj in w.languagelinks]
    else:
        languagelinks = []
    w.languagelinks = []
    text = ET.tostring(e, encoding='utf-8')
    return text, [], languagelinks



from aardtools import xdxf
from StringIO import StringIO
from collections import defaultdict
class Compiler:

    def __init__(self):
        self.articles = defaultdict(list)
        self.redirects = defaultdict(list)

    def add_article(self, title, serialized_article, redirect=False):
        if redirect:
            self.redirects[title].append(serialized_article)
        else:
            self.articles[title].append(serialized_article)

    def add_metadata(self, key, value):
        pass


def test_nu_tag():

    compiler = Compiler()
    parser = xdxf.XDXFParser(compiler)
    xdxf_xml = """<?xml version="1.0" encoding="UTF-8" ?>
<!DOCTYPE xdxf SYSTEM "http://xdxf.sourceforge.net/xdxf_lousy.dtd">
<xdxf lang_from="ENG" lang_to="ENG" format="visual">
<ar><k>abc<nu>|</nu>def</k>
</ar>
</xdxf>
"""
    parser.parse(StringIO(xdxf_xml))
    assert 'abcdef' in compiler.articles

def test_opt_tag():
    compiler = Compiler()
    parser = xdxf.XDXFParser(compiler)
    xdxf_xml = """<?xml version="1.0" encoding="UTF-8" ?>
<!DOCTYPE xdxf SYSTEM "http://xdxf.sourceforge.net/xdxf_lousy.dtd">
<xdxf lang_from="ENG" lang_to="ENG" format="visual">
<ar><k><opt>1</opt>a<opt>2</opt>b<opt>3</opt></k>
</ar>
</xdxf>
"""
    parser.parse(StringIO(xdxf_xml))
    assert 'ab' in compiler.articles
    for s in ('1ab', 'a2b', 'ab3', '1a2b', 'a2b3', '1ab3', '1a2b3'):
        assert s not in compiler.articles
        assert s in compiler.redirects

def test_multiple_k_tags():

    compiler = Compiler()
    parser = xdxf.XDXFParser(compiler)
    xdxf_xml = """<?xml version="1.0" encoding="UTF-8" ?>
<!DOCTYPE xdxf SYSTEM "http://xdxf.sourceforge.net/xdxf_lousy.dtd">
<xdxf lang_from="ENG" lang_to="ENG" format="visual">
<ar><k>a</k>, <k>b</k>, <k>c</k>
</ar>
</xdxf>
"""
    parser.parse(StringIO(xdxf_xml))
    assert 'a' in compiler.articles
    assert 'b' in compiler.redirects
    assert 'c' in compiler.redirects

def test_opt_and_nu_together():

    compiler = Compiler()
    parser = xdxf.XDXFParser(compiler)
    xdxf_xml = """<?xml version="1.0" encoding="UTF-8" ?>
<!DOCTYPE xdxf SYSTEM "http://xdxf.sourceforge.net/xdxf_lousy.dtd">
<xdxf lang_from="ENG" lang_to="ENG" format="visual">
<ar><k>abc<nu>|</nu>def<opt>g</opt></k>
</ar>
</xdxf>
"""
    parser.parse(StringIO(xdxf_xml))
    assert 'abcdef' in compiler.articles
    assert 'abcdefg' in compiler.redirects



"""
This code is based on eqhtml module from `mt-math`_

.. _mt-math: http://www.amk.ca/python/code/mt-math

"""
from __future__ import with_statement
import os
import tempfile
import binascii
import re
import shutil
import xml.etree.ElementTree as etree

from subprocess import Popen, PIPE

latex_doc = r'''\documentclass{article}
\usepackage{amsmath}
\usepackage{amsthm}
\usepackage{amssymb}
\usepackage{bm}
\pagestyle{empty}

%% The following new command definitions and comments are copied
%% from Blahtex(ml) (http://gva.noekeon.org/blahtexml/)
%% for texvc compatibility

%% First we have some macros which are not part of tex/latex/amslatex
%% but which texvc recognises, so for backward compatibility we define
%% them here too. Most of these are apparently intended to cater for
%% those more familiar with HTML entities.

\newcommand{\R}{{\mathbb R}}
\newcommand{\Reals}{\R}
\newcommand{\reals}{\R}
\newcommand{\Z}{{\mathbb Z}}
\newcommand{\N}{{\mathbb N}}
\newcommand{\natnums}{\N}
\newcommand{\Complex}{{\mathbb C}}
\newcommand{\cnums}{\Complex}
\newcommand{\alefsym}{\aleph}
\newcommand{\alef}{\aleph}
\newcommand{\larr}{\leftarrow}
\newcommand{\rarr}{\rightarrow}
\newcommand{\Larr}{\Leftarrow}
\newcommand{\lArr}{\Leftarrow}
\newcommand{\Rarr}{\Rightarrow}
\newcommand{\rArr}{\Rightarrow}
\newcommand{\uarr}{\uparrow}
\newcommand{\uArr}{\Uparrow}
\newcommand{\Uarr}{\Uparrow}
\newcommand{\darr}{\downarrow}
\newcommand{\dArr}{\Downarrow}
\newcommand{\Darr}{\Downarrow}
\newcommand{\lrarr}{\leftrightarrow}
\newcommand{\harr}{\leftrightarrow}
\newcommand{\Lrarr}{\Leftrightarrow}
\newcommand{\Harr}{\Leftrightarrow}
\newcommand{\lrArr}{\Leftrightarrow}
%% The next one looks like a typo in the texvc source code:
\newcommand{\hAar}{\Leftrightarrow}
\newcommand{\sub}{\subset}
\newcommand{\supe}{\supseteq}
\newcommand{\sube}{\subseteq}
\newcommand{\infin}{\infty}
\newcommand{\lang}{\langle}
\newcommand{\rang}{\rangle}
\newcommand{\real}{\Re}
\newcommand{\image}{\Im}
\newcommand{\bull}{\bullet}
\newcommand{\weierp}{\wp}
\newcommand{\isin}{\in}
\newcommand{\plusmn}{\pm}
\newcommand{\Dagger}{\ddagger}
\newcommand{\exist}{\exists}
\newcommand{\sect}{\S}
\newcommand{\clubs}{\clubsuit}
\newcommand{\spades}{\spadesuit}
\newcommand{\hearts}{\heartsuit}
\newcommand{\diamonds}{\diamondsuit}
\newcommand{\sdot}{\cdot}
\newcommand{\ang}{\angle}
\newcommand{\thetasym}{\theta}
\newcommand{\Alpha}{A}
\newcommand{\Beta}{B}
\newcommand{\Epsilon}{E}
\newcommand{\Zeta}{Z}
\newcommand{\Eta}{H}
\newcommand{\Iota}{I}
\newcommand{\Kappa}{K}
\newcommand{\Mu}{M}
\newcommand{\Nu}{N}
\newcommand{\Rho}{P}
\newcommand{\Tau}{T}
\newcommand{\Chi}{X}
\newcommand{\arccot}{\operatorname{arccot}}
\newcommand{\arcsec}{\operatorname{arcsec}}
\newcommand{\arccsc}{\operatorname{arccsc}}
\newcommand{\sgn}{\operatorname{sgn}}

%% The commands in this next group are defined in tex/latex/amslatex,
%% but they don't get mapped to what texvc thinks (e.g. "\part" is used
%% in typesetting books to mean a unit somewhat larger than a chapter,
%% like "Part IV").
%%
%% We'll stick to the way texvc does it, especially since wikipedia has
%% quite a number of equations using them.
\renewcommand{\empty}{\emptyset}
\renewcommand{\and}{\wedge}
\renewcommand{\or}{\vee}
\renewcommand{\part}{\partial}

%% Now we come to the xxxReserved commands. These are all implemented
%% as macros in TeX, so for maximum compatibility, we want to treat
%% their arguments the way a TeX macro does. The strategy is the
%% following. First, in Manager::ProcessInput, we convert e.g. "\mbox"
%% into "\mboxReserved". Then, the MacroProcessor object sees e.g.
%% "\mboxReserved A" and converts it to "\mbox{A}". This simplifies
%% things enormously for the parser, since now it can treat "\mbox"
%% and "\hbox" in the same way. ("\hbox" requires braces around its
%% argument, even if it's just a single character.) This strategy also
%% keeps TeX happy when we send off the purified TeX, since TeX doesn't
%% care about the extra braces.

\newcommand{\mboxReserved}     [1]{\mbox{#1}}
\newcommand{\substackReserved} [1]{\substack{#1}}
\newcommand{\oversetReserved}  [2]{\overset{#1}{#2}}
\newcommand{\undersetReserved} [2]{\underset{#1}{#2}}

%% The following are all similar, but they get extra "safety braces"
%% placed around them. For example, "x^\frac yz" is legal, because it
%% becomes "x^{y \over z}".

\newcommand{\textReserved}     [1]{{\text{#1}}}
\newcommand{\textitReserved}   [1]{{\textit{#1}}}
\newcommand{\textrmReserved}   [1]{{\textrm{#1}}}
\newcommand{\textbfReserved}   [1]{{\textbf{#1}}}
\newcommand{\textsfReserved}   [1]{{\textsf{#1}}}
\newcommand{\textttReserved}   [1]{{\texttt{#1}}}
\newcommand{\emphReserved}     [1]{{\emph{#1}}}
\newcommand{\fracReserved}     [2]{{\frac{#1}{#2}}}
\newcommand{\mathrmReserved}   [1]{{\mathrm{#1}}}
\newcommand{\mathbfReserved}   [1]{{\mathbf{#1}}}
\newcommand{\mathbbReserved}   [1]{{\mathbb{#1}}}
\newcommand{\mathitReserved}   [1]{{\mathit{#1}}}
\newcommand{\mathcalReserved}  [1]{{\mathcal{#1}}}
\newcommand{\mathfrakReserved} [1]{{\mathfrak{#1}}}
\newcommand{\mathttReserved}   [1]{{\mathtt{#1}}}
\newcommand{\mathsfReserved}   [1]{{\mathsf{#1}}}
\newcommand{\bigReserved}      [1]{{\big#1}}
\newcommand{\biggReserved}     [1]{{\bigg#1}}
\newcommand{\BigReserved}      [1]{{\Big#1}}
\newcommand{\BiggReserved}     [1]{{\Bigg#1}}

\newcommand{\japReserved}     [1]{{\jap{#1}}}
\newcommand{\cyrReserved}     [1]{{\cyr{#1}}}

\begin{document}
%s
\end{document}
'''

emptylines = re.compile(r'[\r\n]{2,}')


class MathRenderingFailed(Exception):

    def __init__(self, equation, cmd, error):
        Exception.__init__(self, equation, cmd, error)
        self.equation = equation
        self.cmd = cmd
        self.error = error

    def __str__(self):
        return ("Couldn't convert equation %r (failed cmd: %r, error: %r)"
                % (self.equation, self.cmd, self.error))



def mkpng_texvc(workdir, equation):
    cmd = ['texvc', workdir, workdir, equation, "UTF-8", "72"]
    sub = Popen(cmd, stdout=PIPE, stdin=PIPE, stderr=PIPE)
    (result, error) = sub.communicate()
    if sub.returncode != 0:
        raise MathRenderingFailed(equation, ' '.join(cmd), error)
    else:
        png_fn = os.path.join(workdir, result[1:33] + '.png')
        return png_fn

def mkpng_blahtex(workdir, equation):
    tex_cmd = ['blahtexml', '--texvc-compatible-commands', '--png',
               '--temp-directory', workdir, '--png-directory', workdir]
    sub = Popen(tex_cmd, stdout=PIPE, stdin=PIPE, stderr=PIPE)
    result, error = sub.communicate(equation)
    if sub.returncode != 0:
        raise MathRenderingFailed(equation, ' '.join(tex_cmd), error)
    e = etree.fromstring(result)
    png_fn = e.findtext('png/md5')
    if not png_fn:
        error = e.findtext('error/message')
        raise MathRenderingFailed(equation, ' '.join(tex_cmd), error)
    return os.path.join(workdir, os.path.extsep.join((png_fn, 'png')))

def mkpng_latex(workdir, equation):
    tex_file = os.path.join(workdir, 'eq.tex')

    equation = emptylines.sub('\n', equation)
    eq_stripped = equation.strip().lower()
    if not (eq_stripped.startswith(r'\begin') or
            eq_stripped.startswith('$') or
            eq_stripped.startswith('\\[')):
        equation = '\\[%s\\]' % equation

    doc_text = latex_doc % equation
    with open(tex_file, 'w+') as f:
        f.write(doc_text)

    tex_cmd = ['latex', '-halt-on-error', '-output-directory', workdir, tex_file]
    sub = Popen(tex_cmd, stdout=PIPE, stdin=PIPE, stderr=PIPE)
    error = sub.communicate()[1]
    if sub.returncode != 0:
        raise MathRenderingFailed(equation, ' '.join(tex_cmd), error)

    dvi_file = os.path.join(workdir, 'eq.dvi')
    png_file = os.path.join(workdir, 'eq.png')

    png_cmd = ['dvipng', '-T', 'tight', '-x', '1200', '-z', '9',
               '-bg', 'Transparent', '-o', png_file, dvi_file]

    sub = Popen(png_cmd, stdout=PIPE, stdin=PIPE, stderr=PIPE)
    error = sub.communicate()[1]
    if sub.returncode != 0:
        raise MathRenderingFailed(equation, ' '.join(png_cmd), error)
    return png_file


def toimg(equation, cmd='latex', keeptemp=False):
    try:
        workdir = tempfile.mkdtemp(prefix='math-')

        if isinstance(equation, unicode):
            equation = equation.encode('utf8')

        png_file = globals()['mkpng_'+cmd](workdir, equation)

        with open(png_file, 'rb') as png:
            png_data = png.read()

        imgdata = binascii.b2a_base64(png_data).replace('\n', '')
        return imgdata
    finally:
        if not keeptemp:
            shutil.rmtree(workdir)





# This file is part of Aard Dictionary Tools <http://aarddict.org>.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License <http://www.gnu.org/licenses/gpl-3.0.txt>
# for more details.
#
# Copyright (C) 2008-2009  Igor Tkach

from aarddict import dictionary

def total(inputfile, options):
    d = dictionary.Volume(inputfile)
    d.close()
    return len(d)

def collect_articles(input_file, options, compiler):
    p = AardParser(compiler)
    p.parse(input_file)

def make_input(input_file_name):
    return input_file_name

class AardParser():

    def __init__(self, consumer):
        self.consumer = consumer

    def parse(self, f):
        d = dictionary.Volume(f)
        for key, val in d.metadata.iteritems():
            self.consumer.add_metadata(key, val)
        for i, article in enumerate(d.articles):
            title= d.words[i]
            self.consumer.add_article(title, article)
        d.close()



#!/usr/bin/python

# This file is part of Aard Dictionary Tools <http://aarddict.org>.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License <http://www.gnu.org/licenses/gpl-3.0.txt>
# for more details.
#
# Copyright (C) 2008-2009  Jeremy Mortis, Igor Tkach


from __future__ import with_statement
import uuid
import logging
import sys
import struct
import os
import tempfile
import optparse
import functools
import time
import shutil
from datetime import timedelta

from PyICU import Locale, Collator

try:
    import json
except ImportError:
    import simplejson as json

from aarddict.dictionary import HEADER_SPEC, spec_len, calcsha1, collation_key
import aardtools


log = logging.getLogger('compiler')

tojson = functools.partial(json.dumps, ensure_ascii=False)

MAX_FAT32_FILE_SIZE = 2**32-1

KEY_LENGTH_FORMAT = '>H'
ARTICLE_LENGTH_FORMAT = '>L'
INDEX1_ITEM_FORMAT = '>LL'

def make_opt_parser():
    usage = "Usage: %prog [options] (wiki|xdxf|aard) FILE"
    parser = optparse.OptionParser(version="%prog "+aardtools.__version__, usage=usage)
    parser.add_option(
        '-o', '--output-file',
        default='',
        help=
        'Output file name. By default is the same as input '
        'file base name with .aar extension'
        )
    parser.add_option(
        '-s', '--max-file-size',
        default=str(2**31-1),
        help=
        'Maximum file size in bytes, kilobytes(K), megabytes(M) or gigabytes(G). '
        'Default: %default bytes'
        )
    parser.add_option(
        '-t', '--templates',
        default=None,
        help='Template definitions database'
        )
    parser.add_option(
        '-d', '--debug',
        action='store_true',
        default=False,
        help='Turn on debugging messages'
        )
    parser.add_option(
        '-q', '--quite',
        action='store_true',
        default=False,
        help='Print minimal information about compilation progress'
        )
    parser.add_option(
        '--timeout',
        type='float',
        default=2.0,
        help=
        'Skip article if it was not process in the amount of time '
        'specified. Default: %defaults'
        )
    parser.add_option(
        '--processes',
        type='int',
        default=None,
        help=
        'Size of the worker pool (by default equals to the '
        'number of detected CPUs).'
        )
    parser.add_option(
        '--nomp',
        action='store_true',
        default=False,
        help='Disable multiprocessing, useful for debugging.'
        )
    parser.add_option(
        '--metadata',
        default=None,
        help='INI containing dictionary metadata in [metadata] section'
        )
    parser.add_option(
        '--license',
        default=None,
        help='Name of a UTF-8 encoded text file containing license text'
        )
    parser.add_option(
        '--copyright',
        default=None,
        help='Name of a UTF-8 encoded text file containing copyright notice'
        )
    parser.add_option(
        '--work-dir',
        default='.',
        help=
        'Directory for temporary file created during compilatiod. '
        'Default: %default'
        )
    parser.add_option(
        '--start',
        default=0,
        type='int',
        help='Starting article, skip all articles before. Default: %default'
        )

    parser.add_option(
        '--end',
        default=None,
        type='int',
        help='End article, stop processing at this article. Default: %default'
        )

    parser.add_option(
        '--dict-ver',
        help='Version of the compiled dictionary'
        )

    parser.add_option(
        '--dict-update',
        default='1',
        help='Update number for the compiled dictionary. Default: %default'
        )

    parser.add_option(
        '--wiki-lang',
        help='Wikipedia language (like en, de, fr). This may be different from actual language '
        'in which articles are written. For example, the value for Simple English Wikipedia  is "simple" '
        '(although the actual articles language is "en"). This is inferred from input file name '
        'if it follows same naming pattern as Wiki XML dumps and starts with "{lang}wiki". '
        'Default: %default'
        )

    parser.add_option(
        '--mp-chunk-size',
        default=10000,
        type='int',
        help='This value defines maximum number articles to be processed by pool'
        'of worker processes before it is closed and new pool is created. Typically'
        'there should be no need to change the default value.'
        'Default: %default'
        )

    parser.add_option(
        '--show-legend',
        action='store_true',
        help='Show progress legend'
        )

    parser.add_option('--log-file',
                       help='Log file name. By default derived from output '
                       'file name by adding .log extension')

    parser.add_option('-r', '--remove-session-dir',
                      action='store_true',
                      help='Remove session directory after compilation.')

    parser.add_option(
        '--lang-links',
        help='Add Wikipedia language links to index for these languages '
        '(comma separated list of language codes). Default: %default')

    parser.add_option(
        '--article-count',
        default=0,
        type='int',
        help=('Request specific number of articles, skip redirects '
              '(if set to a number greater then 0). '
              'Default: %default'))

    parser.add_option(
        '--skip-article-title',
        action='store_true',
        help=('Do not include article key in rendered article: '
              'some XDXF dictionaries already inlude title in article text and '
              'needs this to avoid title duplication'))

    parser.add_option('--siteinfo',
                      help='Mediawiki JSON-formatted site info file')


    parser.add_option('--rtl',
                      action="store_true",
                      help='Set direction for Wikipedia articles to rtl')

    return parser

def utf8(func):
    def f(*args, **kwargs):
        newargs = [arg.encode('utf8') if isinstance(arg, unicode) else arg
                   for arg in args]
        newkwargs = {}
        for key, val in kwargs.iteritems():
            newkwargs[key] = (val.encode('utf8')
                              if isinstance(val, unicode) else val)
        return func(*newargs, **newkwargs)
    f.__doc__ = func.__doc__
    f.__name__ = func.__name__
    f.__dict__.update(func.__dict__)
    return f

class Volume(object):

    class ExceedsMaxSize(Exception): pass

    number = 0

    def __init__(self, header_meta_len, max_file_size, work_dir):
        self.header_meta_len = header_meta_len
        self.max_file_size = max_file_size
        self.index1 = tempfile.NamedTemporaryFile(prefix='index1',
                                                  dir=work_dir)
        log.info('Creating temporary index 1 file %s', self.index1.name)
        self.index2 = tempfile.NamedTemporaryFile(prefix='index2',
                                                  dir=work_dir)
        log.info('Creating temporary index 2 file %s', self.index2.name)
        self.articles =  tempfile.NamedTemporaryFile(prefix='articles',
                                                     dir=work_dir)
        log.info('Creating temporary articles file %s', self.articles.name)
        self.index1Length = 0
        self.index2Length = 0
        self.articles_len = 0
        self.index_count = 0
        Volume.number += 1

    def add(self, index1_unit, index2_unit, article_unit):
        if sum((self.header_meta_len,
                self.index1Length,
                self.index2Length,
                self.articles_len,
                len(index1_unit),
                len(index2_unit),
                len(article_unit)
                )) > self.max_file_size:
            raise Volume.ExceedsMaxSize
        self.index1.write(index1_unit)
        self.index1Length += len(index1_unit)
        self.index2.write(index2_unit)
        self.index2Length += len(index2_unit)
        self.index_count += 1
        self.articles.write(article_unit)
        self.articles_len += len(article_unit)


    def flush(self):
        self.index1.flush()
        self.index2.flush()
        self.articles.flush()

    def totuple(self):
        return (self.index1, self.index1Length, self.index2,
                self.index2Length, self.articles, self.articles_len,
                self.index_count)

import threading
article_add_lock = threading.RLock()

class Stats(object):

    def __init__(self):
        self.total = 0
        self.skipped = 0
        self.failed = 0
        self.empty = 0
        self.timedout = 0
        self.articles = 0
        self.redirects = 0
        self.start_time = time.time()

    processed = property(lambda self: (self.articles +
                                       self.redirects +
                                       self.skipped +
                                       self.empty +
                                       self.timedout +
                                       self.failed))


    average = property(lambda self: (self.processed/(time.time() -
                                                     self.start_time)))

    elapsed = property(lambda self: timedelta(seconds=(int(time.time() -
                                                           self.start_time))))

    def __str__(self):
        return ('total: %d, skipped: %d, failed: %d, '
                'empty: %d, timed out: %d, articles: %d, '
                'redirects: %d, average: %.2f/s '
                'elapsed: %s' % (self.total,
                                 self.skipped,
                                 self.failed,
                                 self.empty,
                                 self.timedout,
                                 self.articles,
                                 self.redirects,
                                 self.average,
                                 self.elapsed))


import mmap

class TempArticleStore(object):

    def __init__(self, work_dir=None):
        fd, self.title_store_name = tempfile.mkstemp(prefix='aa-', suffix='.titles', dir=work_dir)
        self.title_store = os.fdopen(fd, 'w')
        fd, self.store_idx_name = tempfile.mkstemp(suffix='.index',
                                                   prefix='aa-', 
                                                   dir=work_dir)
        self.store_idx = os.fdopen(fd, 'wb')

        fd, self.article_store_name = tempfile.mkstemp(suffix='.articles',
                                                        prefix='aa-', 
                                                        dir=work_dir)
        self.article_store = os.fdopen(fd, 'wb')

        self.title_start = 0
        self.article_start = 0
        idx_format = '>IHQI'
        self.pack = functools.partial(struct.pack, idx_format)
        self.unpack = functools.partial(struct.unpack, idx_format)
        self.fmt_size = struct.calcsize(idx_format)

    def append(self, title, article):
        self.title_store.write(title)
        title_len = len(title)        
        
        self.article_store.write(article)
        article_len = len(article)        
        
        self.store_idx.write(self.pack(self.title_start, title_len, 
                                       self.article_start, article_len))

        self.title_start += title_len
        self.article_start += article_len
        

    def sorted(self, key=None):
        """ Return generator that produces ordered (title, article) 
        pairs sorted by title.

        :param key: function of one argument that takes article title 
                    and returns sort key for this title, title itself is used 
                    as key if key function is None        
        """

        self.title_store.flush()
        self.article_store.flush()
        self.store_idx.flush()

        if key is None:
            key = lambda x: x

        with open(self.title_store_name, 'r+') as title_store_f:
            with open(self.article_store_name, 'r+') as article_store_f:
                with open(self.store_idx_name, 'r+b') as store_idx_f:

                    title_store = mmap.mmap(title_store_f.fileno(), 0)
                    article_store = mmap.mmap(article_store_f.fileno(), 0)
                    store_idx = mmap.mmap(store_idx_f.fileno(), 0)

                    def index_item_at(pos):
                        pos_start = pos*self.fmt_size
                        pos_end = pos_start + self.fmt_size
                        return self.unpack(store_idx[pos_start:pos_end])


                    def realkey(x):
                        index_item = index_item_at(x)
                        title_start = index_item[0]
                        title_len = index_item[1]
                        title_end = title_start+title_len
                        title = title_store[title_start:title_end]
                        return key(title)

                    for i in sorted(xrange(len(store_idx)/self.fmt_size),
                                    key=realkey):
                        title_start, title_len, article_start, article_len = index_item_at(i)
                        yield (title_store[title_start:title_start+title_len], 
                               article_store[article_start:article_start+article_len])

    def close(self):
        self.title_store.close()
        self.article_store.close()
        self.store_idx.close()
        os.remove(self.title_store_name)
        os.remove(self.article_store_name)
        os.remove(self.store_idx_name)        

class Compiler(object):

    def __init__(self, output_file_name, max_file_size, session_dir, metadata=None):
        self.uuid = uuid.uuid4()
        self.output_file_name = output_file_name
        self.max_file_size = max_file_size
        self.index_count = 0
        self.session_dir = session_dir
        self.failed_articles = open(os.path.join(self.session_dir, "failed.txt"), 'w')
        self.empty_articles = open(os.path.join(self.session_dir, "empty.txt"), 'w')
        self.skipped_articles = open(os.path.join(self.session_dir, "skipped.txt"), 'w')
        self.metadata = metadata if metadata is not None else {}
        self.file_names = []
        self.stats = Stats()
        self.last_stat_update = 0
        self.article_store = TempArticleStore(self.session_dir)
        log.info('Collecting articles')

    def add_metadata(self, key, value):
        if key not in self.metadata:
            self.metadata[key] = value
        else:
            log.warn('Value for metadata key %s is already set, '
                     'new value %s will be ignored',
                     key, value)

    @utf8
    def add_article(self, title, serialized_article, redirect=False, count=True):
        with article_add_lock:
            if not title:
                log.warn('Blank title, ignoring article "%s"',
                         serialized_article)
                return
            if not serialized_article:
                self.empty_article(title)
                return
            log.debug('Adding article for "%s"', title)
            self.article_store.append(title, compress(serialized_article))
            if count:
                if not redirect:
                    self.stats.articles += 1
                else:
                    self.stats.redirects += 1
            self.print_stats()

    @utf8
    def fail_article(self, title):
        self.stats.failed += 1
        self.failed_articles.write(title+'\n')
        self.print_stats()

    @utf8
    def empty_article(self, title):
        self.stats.empty += 1
        self.empty_articles.write(title+'\n')
        self.print_stats()

    @utf8
    def skip_article(self, title):
        self.stats.skipped += 1
        self.skipped_articles.write(title+'\n')
        self.print_stats()

    def timedout(self, count=1):
        self.stats.timedout += count
        self.print_stats()

    def print_stats(self):
        t = time.time()
        if (t - self.last_stat_update) > 1.0:
            self.last_stat_update = t
            print_progress(self.stats)

    def compile(self):
        print_progress(self.stats)
        writeln()
        self.failed_articles.close()
        self.empty_articles.close()
        self.skipped_articles.close()
        writeln('Compiling .aar files')
        self.add_metadata("article_count", self.stats.articles)
        articles = self.article_store.sorted(key=lambda x:
                                                 collation_key(x).getByteArray())
        log.info('Compiling %s', self.output_file_name)
        metadata = compress(tojson(self.metadata).encode('utf8'))
        header_meta_len = spec_len(HEADER_SPEC) + len(metadata)
        create_volume_func = functools.partial(self.create_volume,
                                               header_meta_len)
        for volume in self.make_volumes(create_volume_func, articles):
            m = "Creating volume %d" % volume.number
            log.info(m)
            writeln(m).flush()
            file_name = self.make_aar(volume)
            self.file_names.append(file_name)
            m = "Wrote volume %d" % volume.number
            log.info(m)
            writeln(m).flush()
        self.article_store.close()
        self.write_volume_count()
        self.write_sha1sum()
        rename_files(self.file_names)

    def create_volume(self, header_meta_len):
        return Volume(header_meta_len, self.max_file_size, self.session_dir)

    def make_volumes(self, create_volume_func, articles):
        volume = create_volume_func()
        for title, serialized_article in articles:
            index1Unit = struct.pack(INDEX1_ITEM_FORMAT,
                                     volume.index2Length,
                                     volume.articles_len)
            index2Unit = struct.pack(KEY_LENGTH_FORMAT, len(title)) + title
            article_unit = (struct.pack(ARTICLE_LENGTH_FORMAT,
                                       len(serialized_article)) +
                            serialized_article)
            try:
                volume.add(index1Unit, index2Unit, article_unit)
            except Volume.ExceedsMaxSize:
                volume.flush()
                yield volume
                volume = create_volume_func()
                index1Unit = struct.pack(INDEX1_ITEM_FORMAT,
                                         volume.index2Length,
                                         volume.articles_len)
                volume.add(index1Unit, index2Unit, article_unit)
        volume.flush()
        yield volume

    def write_header(self, output_file, meta_length, index1Length,
                     index2Length, index_count, volume):
        article_offset = (spec_len(HEADER_SPEC) + meta_length +
                          index1Length + index2Length)
        values = dict(signature='aard',
                      sha1sum='0'*40,
                      version=1,
                      uuid=self.uuid.bytes,
                      volume=volume,
                      of=0,
                      total_volumes=0,
                      meta_length=meta_length,
                      index_count=index_count,
                      article_offset=article_offset,
                      index1_item_format=INDEX1_ITEM_FORMAT,
                      key_length_format=KEY_LENGTH_FORMAT,
                      article_length_format=ARTICLE_LENGTH_FORMAT)
        for name, fmt in HEADER_SPEC:
            output_file.write(struct.pack(fmt, values[name]))

    def write_meta(self, output_file, metadata):
        output_file.write(metadata)

    def write_index1(self, output_file, index1):
        log.debug('Writing index 1')
        index1.seek(0)
        count = 0
        while True:
            unit = index1.read(struct.calcsize(INDEX1_ITEM_FORMAT))
            if len(unit) == 0:
                break
            index2ptr, offset = struct.unpack(INDEX1_ITEM_FORMAT, unit)
            unit = struct.pack(INDEX1_ITEM_FORMAT, index2ptr, offset)
            output_file.write(unit)
            count += 1
        log.debug('Wrote %d items to index 1', count)
        index1.close()

    def write_index2(self, output_file, index2):
        log.debug('Writing index 2')
        index2.seek(0)
        count = 0
        while True:
            unitLengthString = index2.read(struct.calcsize(KEY_LENGTH_FORMAT))
            if len(unitLengthString) == 0:
                break
            count += 1
            unitLength, = struct.unpack(KEY_LENGTH_FORMAT, unitLengthString)
            unit = index2.read(unitLength)
            output_file.write(unitLengthString + unit)
        log.debug('Wrote %d items to index 2', count)
        index2.close()

    def write_articles(self, output_file, articles):
        articles.seek(0)
        count = 0
        while True:
            unitLengthString = articles.read(struct.
                                             calcsize(ARTICLE_LENGTH_FORMAT))
            if len(unitLengthString) == 0:
                break
            count += 1
            unitLength, = struct.unpack(ARTICLE_LENGTH_FORMAT,
                                        unitLengthString)
            unit = articles.read(unitLength)
            output_file.write(unitLengthString + unit)
        log.debug('Wrote %d items to articles', count)
        articles.close()

    def write_sha1sum(self):
        for file_name in self.file_names:
            msg = "Calculating checksum for %s" % file_name
            log.info(msg)
            offset = spec_len(HEADER_SPEC[:2])
            st_size = os.stat(file_name).st_size
            size = float(st_size - offset)
            for pos, sha1sum in calcsha1(file_name, offset):
                (display.erase_line().cr()
                .write(msg).write(': ').write('%.1f%%' % (100*pos/size)))
            sha1sum = sha1sum.hexdigest()
            msg = "%s sha1: %s" % (file_name, sha1sum)
            log.info(msg)
            display.erase_line().cr().writeln(msg)
            output_file = open(file_name, "r+b")
            output_file.seek(spec_len(HEADER_SPEC[:1]))
            output_file.write(sha1sum)
            output_file.close()

    def make_aar(self, volume):
        (index1, index1Length, index2, index2Length, articles,
         articles_len, index_count) = volume.totuple()
        file_name = '%s.%d' % (self.output_file_name, Volume.number)
        output_file = open(file_name, "wb", 8192)
        metadata = compress(tojson(self.metadata).encode('utf8'))
        self.write_header(output_file, len(metadata), index1Length,
                          index2Length, index_count, Volume.number)
        self.write_meta(output_file, metadata)
        self.write_index1(output_file, index1)
        self.write_index2(output_file, index2)
        self.write_articles(output_file, articles)
        output_file.close()
        log.info("Done with %s", file_name)
        return file_name

    def write_volume_count(self):
        name, fmt = HEADER_SPEC[5]
        log.info("Writing volume count %d to all volumes as %s",
                 Volume.number, fmt)
        log.debug('Writing' )
        for file_name in self.file_names:
            output_file = open(file_name, "r+b")
            output_file.seek(spec_len(HEADER_SPEC[:5]))
            output_file.write(struct.pack(fmt, Volume.number))
            output_file.close()

def rename_files(file_names):
    """
    >>> from minimock import mock
    >>> import compiler
    >>> mock('compiler.rename_file', returns_func=lambda f, p, args: None)
    >>> Volume.number = 2
    >>> rename_files(['enwiki-20090530-2.aar.1'])
    Called compiler.rename_file(
        'enwiki-20090530-2.aar.1',
        '%s.%s',
        ('enwiki-20090530-2', 'aar'))

    >>> rename_files(['enwiki-20090530-2.aar.1', 'enwiki-20090530-2.aar.2'])
    Called compiler.rename_file(
        'enwiki-20090530-2.aar.1',
        '%s.%s_of_%s.%s',
        ('enwiki-20090530-2', '1', 2, 'aar'))
    Called compiler.rename_file(
        'enwiki-20090530-2.aar.2',
        '%s.%s_of_%s.%s',
        ('enwiki-20090530-2', '2', 2, 'aar'))

    >>> rename_files(['enwiki-20090530-2.1', 'enwiki-20090530-2.2'])
    Called compiler.rename_file(
        'enwiki-20090530-2.1',
        '%s.%s_of_%s.%s',
        ('enwiki-20090530-2', '1', 2, 'aar'))
    Called compiler.rename_file(
        'enwiki-20090530-2.2',
        '%s.%s_of_%s.%s',
        ('enwiki-20090530-2', '2', 2, 'aar'))

    """
    one = len(file_names) == 1
    pattern = '%s.%s' if one else '%s.%s_of_%s.%s'
    for file_name in file_names:
        if file_name.count('.') == 1:
            base, vol = file_name.split('.')
            ext = 'aar'
        else:
            base, ext, vol = file_name.rsplit('.', 2)
        args = (base, ext) if one else (base, vol, Volume.number, ext)
        rename_file(file_name, pattern, args)

def rename_file(file_name, newname_pattern, args):
    newname = newname_pattern % args
    log.info('Renaming %s ==> %s', file_name, newname)
    display.write('Created ').bold(newname).writeln()
    os.rename(file_name, newname)


import zlib
import bz2

def _zlib(s):
    return zlib.compress(s)

def _bz2(s):
    return bz2.compress(s)

from collections import defaultdict
compress_counts = defaultdict(int)

def compress(text):
    compressed = text
    cfunc = None
    for func in (_zlib, _bz2):
        c = func(text)
        if len(c) < len(compressed):
            compressed = c
            cfunc = func
    if cfunc:
        compress_counts[cfunc.__name__] += 1
    else:
        compress_counts['none'] += 1
    return compressed


collator = Collator.createInstance(Locale(''))
collator.setStrength(Collator.QUATERNARY)
collation_key = collator.getCollationKey


def make_output_file_name(input_file, options):
    """
    Return output file name based on input file name.

    >>> from minimock import Mock
    >>> opts = Mock('options')
    >>> opts.output_file = 'abc'
    >>> make_output_file_name('123.tar.bz2', opts)
    'abc'

    >>> opts.output_file = None
    >>> make_output_file_name('123.tar.bz2', opts)
    '123.aar'

    >>> make_output_file_name('-', opts)
    'dictionary.aar'

    """
    if options.output_file:
        output_file = options.output_file
    elif input_file == '-':
        output_file = 'dictionary.aar'
    else:
        output_file = strip_ext(input_file.rstrip(os.path.sep))
        output_file += '.aar'
    return output_file

def strip_ext(fname):
    """
    Return file name with one or two extension stripped
    (depending on extension).

    >>> strip_ext('abc.def.txt')
    'abc.def'
    >>> strip_ext('abc.def.tar.bz2')
    'abc.def'
    >>> strip_ext('abc.def.tar.gz')
    'abc.def'
    >>> strip_ext('abc.def.xml.bz2')
    'abc.def'
    >>> strip_ext('abc.def.xdxf')
    'abc.def'
    >>> strip_ext('/a/b/c/a.cdb/')
    'a'

    """
    output_file = os.path.basename(fname.rstrip(os.path.sep))
    output_file = output_file[:output_file.rfind('.')]
    if (output_file.endswith('.tar') or
        output_file.endswith('.xml') or
        output_file.endswith('.xdxf')):
        output_file = output_file[:output_file.rfind('.')]
    return output_file


def max_file_size(options):
    s = options.max_file_size
    return parse_size(s)

def parse_size(s):
    if s.endswith('M'):
        return int(s.strip('M'))*1024*1024
    elif s.endswith('G'):
        return int(s.strip('G'))**1024*1024*1024
    elif s.endswith('K'):
        return int(s.strip('K'))*1024
    elif s.endswith('m'):
        return int(s.strip('m'))*1000*1000
    elif s.endswith('g'):
        return int(s.strip('g'))*1000*1000*1000
    elif s.endswith('k'):
        return int(s.strip('k'))*1000
    elif s.endswith('b'):
        return int(s.strip('b'))
    else:
        return int(s)

class Display:

    ERASE_LINE = '\033[2K'
    BOLD='\033[1m'
    RED = '\033[91m'
    YELLOW = '\033[93m'
    GREEN = '\033[92m'
    ENDC = '\033[0m'

    def reset_att(self):
        sys.stdout.write(self.ENDC)
        return self

    def ok(self, text):
        sys.stdout.write(self.GREEN + text + self.ENDC)
        return self

    def warn(self, text):
        sys.stdout.write(self.YELLOW + text + self.ENDC)
        return self

    def fail(self, text):
        sys.stdout.write(self.RED + text + self.ENDC)
        return self

    def bold(self, text):
        sys.stdout.write(self.BOLD + text + self.ENDC)
        return self

    def erase_line(self):
        sys.stdout.write(self.ERASE_LINE)
        return self

    def write(self, text):
        sys.stdout.write(text)
        return self

    def writeln(self, text=''):
        self.write(text)
        sys.stdout.write('\n')
        return self

    def cr(self):
        sys.stdout.write('\r')
        return self

    def flush(self):
        sys.stdout.flush()
        return self

display = Display()
writeln = display.writeln

def print_legend():
    (display
    .bold('t').writeln(' - time elapsed')
    .bold('avg').writeln(' - average number of articles processed per second')
    .ok('a').writeln(' - number of processed articles')
    .ok('r').writeln(' - number of processed redirects')
    .warn('s').writeln(' - number of skipped articles')
    .warn('e').writeln(' - number of articles with no text (empty)')
    .fail('to').writeln(' - approximate number of articles that couldn\'t be converted fast enough (timed out)')
    .fail('f').writeln(' - number of articles that couldn\'t be converted (failed)'))


def print_progress(stats):
    try:
        progress = '%.2f' % (100*float(stats.processed)/stats.total) if stats.total else '?'
        (display
         .erase_line()
         .bold('%s%% ' % progress)
         .bold('t: %s ' % stats.elapsed)
         .bold('avg: %.1f/s ' % stats.average)
         .ok('a: %d r: %d ' % (stats.articles, stats.redirects))
         .warn('s: %d ' % stats.skipped)
         .warn('e: %d ' % stats.empty)
         .fail('to: %d ' % stats.timedout)
         .fail('f: %d ' % stats.failed)
         .cr().flush())
    except KeyboardInterrupt:
        display.reset_att()


def guess_version(input_file_name):
    """ Guess dictionary version from input file name.

    >>> guess_version('simplewiki-20090506-pages-articles.cdb')
    '20090506'

    >>> guess_version('~/wikidumps/simplewiki-20090506-pages-articles.cdb/')
    '20090506'

    >>> guess_version('~/wikidumps/simplewiki-20090506-pages-articles.cdb')
    '20090506'

    >>> guess_version('some-name')

    >>> guess_version('ruwiktionary-20090122-pages-articles.cdb')
    '20090122'

    >>> guess_version('ruwiktionary-20090122-1.cdb')
    '20090122'

    >>> guess_version('ruwiktionary-20090122.cdb')
    '20090122'

    """
    import re
    m = re.match(r'\w+-?(\d+)[^\d]*',
                 os.path.basename(input_file_name.rstrip(os.path.sep)))
    return m.group(1) if m else None

def guess_wiki_lang(input_file_name):
    """ Guess wiki language from input file name.

    >>> guess_wiki_lang('simplewiki-20090506-pages-articles.cdb')
    'simple'
    >>> guess_wiki_lang('simplewiki-20090506-pages-articles.cdb/')
    'simple'
    >>> guess_wiki_lang('~/wikidumps/simplewiki-20090506-pages-articles.cdb/')
    'simple'
    >>> guess_wiki_lang('~/wikidumps/simplewiki-20090506-pages-articles.cdb')
    'simple'
    >>> guess_wiki_lang('elwiki-20090512-pages-articles')
    'el'
    >>> guess_wiki_lang('somename')

    >>> guess_wiki_lang('ruwiktionary-20090122-pages-articles')
    'ru'

    """
    import re
    m = re.match(r'([a-zA-Z]{2,})wik[i|t].*',
                 os.path.basename(input_file_name.rstrip(os.path.sep)))
    return m.group(1) if m else None


def main():

    opt_parser = make_opt_parser()
    options, args = opt_parser.parse_args()

    if not args:
        opt_parser.print_help()
        raise SystemExit(1)

    if len(args) < 2:
        sys.stderr.write('Not enough parameters\n')
        opt_parser.print_help()
        raise SystemExit(1)

    input_type = args[0]
    input_files = args[1:]

    if not input_files:
        sys.stderr.write('No input files specified\n')
        raise SystemExit(1)

    if '-' in input_files and len(input_files) != 1:
        sys.stderr.write('stdin is specified as input file, but other files '
                         'are specified too (%s), can\'t proceed\n' % input_files)
        raise SystemExit(1)

    for input_file in input_files:
        if not input_file == '-' and not os.path.exists(input_file):
            sys.stderr.write('No such file: %s\n' % input_file)
            raise SystemExit(1)

    session_dir = os.path.join(options.work_dir,
                               'aardc-'+('%.2f' % time.time()).replace('.','-'))

    if os.path.exists(session_dir):
        sys.stderr.write('Session directory %s already'
                         ' exists, can\'t proceed\n' % session_dir)
        raise SystemExit(1)
    else:
        os.mkdir(session_dir)
        display.write('Session dir ').bold(session_dir).writeln()


    try:
        converter = __import__(input_type, globals=globals())
    except ImportError:
        sys.stderr.write('Unknown input type %s\n' % args[0])
        opt_parser.print_help()
        raise SystemExit(1)

    output_file_name = make_output_file_name(input_files[0], options)

    if options.quite:
        log_level = logging.ERROR
    elif options.debug:
        log_level = logging.DEBUG
    else:
        log_level = logging.INFO

    if options.log_file:
        log_file_name = options.log_file
    else:
        log_file_name = os.path.join(session_dir, 'log')

    display.write('Writing log to ').bold(log_file_name).writeln()
    root_logger = logging.getLogger()
    root_logger.handlers[:] = []    
    logging.basicConfig(format='%(asctime)s %(levelname)s [%(name)s] %(message)s',
                        level=log_level,
                        filename=log_file_name,
                        datefmt="%X")
    multiprocessing_logger = logging.getLogger('multiprocessing')
    #multiprocessing is noisy at info level
    multiprocessing_logger.setLevel(logging.WARNING)
    multiprocessing_logger.handlers = root_logger.handlers

    max_volume_size = max_file_size(options)
    log.info('Maximum file size is %d bytes', max_volume_size)
    if max_volume_size > MAX_FAT32_FILE_SIZE:
        global INDEX1_ITEM_FORMAT
        INDEX1_ITEM_FORMAT = '>LQ'
        log.info('Maximum file size is too big 32-bit offsets, '
                 'setting index item format to %s',
                 INDEX1_ITEM_FORMAT)

    if input_type=='wiki':
        if not options.wiki_lang:
            options.wiki_lang = guess_wiki_lang(input_files[0])
            if not options.wiki_lang:
                sys.stderr.write('Wiki language is neither specified with --wiki-lang '
                                 'not could be guessed from input file name\n')
                raise SystemExit(1)
        log.info('Wikipedia language: %s', options.wiki_lang)

    if not options.dict_ver:
        options.dict_ver = guess_version(input_files[0])
        if options.dict_ver:
            log.info('Using %s as dictionary version', options.dict_ver)
        else:
            options.dict_ver = time.strftime('%Y%m%d%H%M%S')
            log.warn('Dictionary version is not specified and couldn\'t '
                     'be guessed from input file name, using %s', options.dict_ver)

    metadata = {aardtools.__name__: aardtools.__version__}

    if options.license:
        with open(options.license) as f:
            metadata['license'] = f.read()

    if options.copyright:
        with open(options.copyright) as f:
            metadata['copyright'] = f.read()

    log.debug('Metadata: %s', metadata)


    compiler = Compiler(output_file_name, max_volume_size,
                        session_dir, metadata)


    t0 = time.time()
    display.write('Converting ').bold(', '.join(input_files)).writeln()

    if hasattr(converter, 'total'):
        display.write('Calculating total number of articles...').cr().flush()
        for input_file in input_files:
            compiler.stats.total += converter.total(converter.make_input(input_file), options)
    display.erase_line().writeln('total: %d' % compiler.stats.total)

    if options.show_legend:
        print_legend()

    for input_file in input_files:
        log.info('Collecting articles in %s', input_file)
        converter.collect_articles(converter.make_input(input_file), options, compiler)
    compiler.compile()
    if options.remove_session_dir:
        writeln('Removing session dir')
        shutil.rmtree(session_dir)
    log.info(compiler.stats)
    log.info('Compression: %s',
             ', '.join('%s - %d' % item
                      for item in compress_counts.iteritems()))
    log.info('Compilation took %s', timedelta(seconds=time.time() - t0))
    writeln('Compilation took %s' % timedelta(seconds=int(time.time() - t0)))


if __name__ == '__main__':
    main()



import random
import string
from aardtools.compiler import TempArticleStore

def setup():
    global store, data
    store = TempArticleStore()
    random_words = [random_word() for i in range(100)]
    data = [(word, 'article %s' % word) for word in random_words]
    for title, article in data:
        store.append(title, article)

def random_word():
    length = random.randint(1, 20)
    letters = list()
    for i in range(length):
        letters.append(random.choice(string.letters))
    return ''.join(letters)

def teardown():
    store.close()

def test_without_key():
    assert list(store.sorted()) == sorted(data, key=lambda x: x[0])

def test_with_key():
    actual = list(store.sorted(key=lambda x: ''.join(reversed(x))))
    expected = sorted(data, key=lambda x: ''.join(reversed(x[0])))
    assert actual == expected, 'actual:\n%r\nexpected:\n%r\n' % (actual, expected)



#! /usr/bin/env python

import sys
import urllib
try:
    import simplejson as json
except ImportError:
    import json

def fetch(sitehostname):
    url = 'http://%s/w/api.php?action=query&meta=siteinfo&siprop=general|namespaces|namespacealiases|magicwords|interwikimap&format=json' % sitehostname
    sys.stderr.write('fetching %r\n' % url)
    data = urllib.urlopen(url).read()
    data = json.loads(data)['query']
    return json.dumps(data, indent=4, sort_keys=True)

def main():
    argv = sys.argv
    if len(argv) < 2:
        sys.exit('Usage: %s SITEHOSTNAME' % argv[0])

    sitehostname = argv[1]
    serialized_data = fetch(sitehostname)
    sys.stdout.write(serialized_data)

if __name__ == '__main__':
    main()



# This file is part of Aard Dictionary Tools <http://aarddict.org>.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License <http://www.gnu.org/licenses/gpl-3.0.txt>
# for more details.
#
# Copyright (C) 2008-2009  Igor Tkach

import os
import json
import re
import mmap

from collections import defaultdict

#original expression from
#http://stackoverflow.com/questions/694344/regular-expression-that-matches-between-quotes-containing-escaped-quotes
#"(?:[^\\"]+|\\.)*"
#some examples don't have closing quote which
#make the subn with this expression hang
#quoted_text = re.compile(r'"(?:[^"]+|\.)*["|\n]')

#make it a capturing group so that we can get rid of quotes
quoted_text = re.compile(r'"([^"]+|\.)*["|\n]')

ref = re.compile(r"`(\w+)'")

wordnet = None

def total(inputfile, options):
    global wordnet
    wordnet = WordNet(inputfile)
    wordnet.prepare()
    count = 0
    for title in wordnet.collector:
        has_article = False
        for piece in wordnet.collector[title]:
            if isinstance(piece, tuple):
                count += 1
            else:
                has_article = True
        if has_article:
            count += 1
    return count


def collect_articles(input_file, options, compiler):
    wordnet.process(compiler)

def make_input(input_file_name):
    return input_file_name #this should be wordnet dir, leave it alone

def iterlines(wordnetdir):
    dict_dir = os.path.join(wordnetdir, 'dict')
    for name in os.listdir(dict_dir):
        if name.startswith('data.'):
            with open(os.path.join(dict_dir, name)) as f:
                for line in f:
                    if not line.startswith('  '):
                        yield line

class SynSet(object):

    def __init__(self, line):
        self.line = line
        meta, self.gloss = line.split('|')
        self.meta_parts = meta.split()

    @property
    def offset(self):
        return int(self.meta_parts[0])

    @property
    def lex_filenum(self):
        return self.meta_parts[1]

    @property
    def ss_type(self):
        return self.meta_parts[2]

    @property
    def w_cnt(self):
        return int(self.meta_parts[3], 16)

    @property
    def words(self):
        return [self.meta_parts[4+2*i].replace('_', ' ')
                for i in range(self.w_cnt)]

    @property
    def pointers(self):
        p_cnt_index = 4+2*self.w_cnt
        p_cnt = self.meta_parts[p_cnt_index]
        pointer_count = int(p_cnt)
        start = p_cnt_index + 1
        return [Pointer(*self.meta_parts[start+i*4:start+(i+1)*4])
                for i in range(pointer_count)]

    def __repr__(self):
        return 'SynSet(%r)' % self.line


class PointerSymbols(object):

    n = {'!':    'Antonyms',
         '@':    'Hypernyms',
         '@i':   'Instance hypernyms',
         '~':    'Hyponyms',
         '~i':   'Instance hyponyms',
         '#m':   'Member holonyms',
         '#s':   'Substance holonyms',
         '#p':   'Part holonyms',
         '%m':   'Member meronyms',
         '%s':   'Substance meronyms',
         '%p':   'Part meronyms',
         '=':    'Attributes',
         '+':    'Derivationally related forms',
         ';c':   'Domain of synset - TOPIC',
         '-c':   'Member of this domain - TOPIC',
         ';r':   'Domain of synset - REGION',
         '-r':   'Member of this domain - REGION',
         ';u':   'Domain of synset - USAGE',
         '-u':   'Member of this domain - USAGE'}

    v = {'!':   'Antonyms',
         '@':   'Hypernyms',
         '~':   'Hyponyms',
         '*':   'Entailments',
         '>':   'Cause',
         '^':   'Also see',
         '$':   'Verb group',
         '+':   'Derivationally related forms',
         ';c':  'Domain of synset - TOPIC',
         ';r':  'Domain of synset - REGION',
         ';u':  'Domain of synset - USAGE'}

    a = s = {'!':   'Antonyms',
             '+':   'Derivationally related forms',
             '&':   'Similar to',
             '<':   'Participle of verb',
             '\\':  'Pertainyms',
             '=':   'Attributes',
             '^':   'Also see',
             ';c':  'Domain of synset - TOPIC',
             ';r':  'Domain of synset - REGION',
             ';u':  'Domain of synset - USAGE'}

    r = {'!':   'Antonyms',
         '\\':  'Derived from adjective',
         '+':   'Derivationally related forms',
         ';c':  'Domain of synset - TOPIC',
         ';r':  'Domain of synset - REGION',
         ';u':  'Domain of synset - USAGE'}


class Pointer(object):

    def __init__(self, symbol, offset, pos, source_target):
        self.symbol = symbol
        self.offset = int(offset)
        self.pos = pos
        self.source_target = source_target
        self.source = int(source_target[:2], 16)
        self.target = int(source_target[2:], 16)

    def __repr__(self):
        return ('Pointer(%r, %r, %r, %r)' %
                (self.symbol, self.offset,
                 self.pos, self.source_target))


class WordNet():

    def __init__(self, wordnetdir):
        self.wordnetdir = wordnetdir
        self.collector = defaultdict(list)

    def prepare(self):

        ss_types = {'n': 'n.',
                    'v': 'v.',
                    'a': 'adj.',
                    's': 'adj. satellite',
                    'r': 'adv.'}

        file2pos = {'data.adj': ['a', 's'],
                    'data.adv': ['r'],
                    'data.noun': ['n'],
                    'data.verb': ['v']}

        dict_dir = os.path.join(self.wordnetdir, 'dict')

        mmap_files = {}
        for name in os.listdir(dict_dir):
            if name.startswith('data.'):
                if name in file2pos:
                    f = open(os.path.join(dict_dir, name), 'r+')
                    m = mmap.mmap(f.fileno(), 0)
                    for key in file2pos[name]:
                        mmap_files[key] = m

        def a(word):
            return '<a href="%s">%s</a>' % (word, word)

        for line in iterlines(self.wordnetdir):
            synset = SynSet(line)
            gloss_with_examples, _ = quoted_text.subn(lambda x: '<cite class="ex">%s</cite>' %
                                                   x.group(1), synset.gloss)
            gloss_with_examples, _ = ref.subn(lambda x: a(x.group(1)), gloss_with_examples)

            words = synset.words
            for i, word in enumerate(words):
                synonyms = [w for w in words if w != word]
                synonyms_str = ('<br/><small class="co">Synonyms:</small> %s' %
                                ', '.join([a(w) for w in synonyms]) if synonyms else '')
                pointers = defaultdict(list)
                for pointer in synset.pointers:
                    if (pointer.source and pointer.target and
                        pointer.source - 1 != i):
                        continue
                    symbol = pointer.symbol
                    if symbol and symbol[:1] in (';', '-'):
                        continue
                    try:
                        symbol_desc = getattr(PointerSymbols, synset.ss_type)[symbol]
                    except KeyError:
                        print 'WARNING: unknown pointer symbol %s for %s ' % (symbol, synset.ss_type)
                        symbol_desc = symbol

                    mmap_file = mmap_files[pointer.pos]
                    mmap_file.seek(pointer.offset)
                    referenced_synset = SynSet(mmap_file.readline())
                    if pointer.source == 0 and pointer.target == 0:
                        pointers[symbol_desc] = [w for w in referenced_synset.words
                                                 if w not in words]
                    else:
                        referenced_word = referenced_synset.words[pointer.target - 1]
                        if referenced_word not in pointers[symbol_desc]:
                            pointers[symbol_desc].append(referenced_word)

                pointers_str = ''
                for symbol_desc, referenced_words in pointers.iteritems():
                    if referenced_words:
                        pointers_str += '<br/><small class="co">%s:</small> ' % symbol_desc
                        pointers_str += ', '.join([a(w) for w in referenced_words])
                self.collector[word].append('<i class="pos">%s</i> %s%s%s' %
                                            (ss_types[synset.ss_type],
                                             gloss_with_examples,
                                             synonyms_str,
                                             pointers_str))



    def process(self, consumer):

        readme_file = os.path.join(self.wordnetdir, 'README')
        license_file = os.path.join(self.wordnetdir, 'LICENSE')

        consumer.add_metadata('title', 'WordNet')
        consumer.add_metadata('index_language', 'en')
        consumer.add_metadata('article_language', 'en')
        consumer.add_metadata('source', 'http://wordnet.princeton.edu')

        with open(readme_file) as f:
            readme = f.read()
            lines = readme.splitlines()
            lines = lines[5:]
            first_p_index = lines.index('')
            first_p = ' '.join(lines[:first_p_index])
            lines = lines[first_p_index+1:]
            second_p_index = lines.index('')
            second_p = ' '.join(lines[:second_p_index])
            aard_p = ('WordNet for Aard Dictionary is a collection of articles '
                      'consisting of all meanings of a given word and links '
                      'to lexically and sematically related words.')
            consumer.add_metadata('description', 
                                  '\n\n'.join((first_p, second_p, aard_p)))

        with open(license_file) as f:
            license_text = f.read()
            aard_p = """(This is original WordNet license text. 
Note that the software covered by this license 
is WordNet software, not Aard Dictionary.)
"""
            consumer.add_metadata('license', '\n'.join((aard_p, license_text)))
            version = license_text.splitlines()[0].split()[-1]
            consumer.add_metadata('version', version)

        article_template = '<h1>%s</h1><span>%s</span>'

        for title in self.collector:
            pieces = self.collector[title]
            article_pieces = []
            redirects = []
            for piece in pieces:
                if isinstance(piece, tuple):
                    redirects.append(piece)
                else:
                    article_pieces.append(piece)

            article_pieces_count = len(article_pieces)

            text = None
            if article_pieces_count > 1:
                ol = ['<ol>'] + ['<li>%s</li>' % ap for ap in article_pieces] + ['</ol>']
                text = (article_template % (title, ''.join(ol)))
            elif article_pieces_count == 1:
                text = (article_template %
                        (title, article_pieces[0]))

            if text:
                consumer.add_article(title,
                                     json.dumps((text, [])),
                                     redirect=False)

            #add redirects after articles so that
            #redirects to titles that have both articles and
            #redirects land on articles
            for redirect in redirects:
                consumer.add_article(title,
                                     json.dumps(redirect),
                                     redirect=True)



# -*- coding: utf-8 -*-
# This file is part of Aard Dictionary Tools <http://aarddict.org>.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License <http://www.gnu.org/licenses/gpl-3.0.txt>
# for more details.
#
# Copyright (C) 2008-2009  Igor Tkach

from __future__ import with_statement
import functools
import logging
import os
from itertools import islice

try:
    import json
except ImportError:
    import simplejson as json

from mwlib import uparser, xhtmlwriter
from mwlib.log import Log
Log.logfile = None

from mwlib import lrucache, expr
expr._cache = lrucache.mt_lrucache(100)

from mwlib.templ.evaluate import Expander
Expander.parsedTemplateCache = lrucache.lrucache(100)

tojson = functools.partial(json.dumps, ensure_ascii=False)

import multiprocessing
from multiprocessing import Pool, TimeoutError
from mwlib.cdbwiki import WikiDB
from mwlib._version import version as mwlib_version
import mwlib.siteinfo

import gc

import mwaardhtmlwriter as writer

lic_dir = os.path.join(os.path.dirname(__file__), 'licenses')

known_licenses = {"Creative Commons Attribution-Share Alike 3.0 Unported": 
                  os.path.join(lic_dir, "ccasau-3.0.txt"),
                  "GNU Free Documentation License 1.2": 
                  os.path.join(lic_dir, "gfdl-1.2.txt")}

wikidb = None
log = logging.getLogger('wiki')

def _create_wikidb(cdbdir, lang, rtl):
    global wikidb
    wikidb = Wiki(cdbdir, lang, rtl)

def _init_process(cdbdir, lang, rtl):
    global log
    log = multiprocessing.get_logger()
    _create_wikidb(cdbdir, lang, rtl)

class ConvertError(Exception):

    def __init__(self, title):
        Exception.__init__(self, title)
        self.title = title

    def __str__(self):
        """
        >>> e = ConvertError('абвгд'.decode('utf8'))
        >>> print str(e)
        ConvertError: абвгд

        """
        return 'ConvertError: %s' % self.title.encode('utf8')

    def __repr__(self):
        """
        >>> e = ConvertError('абвгд'.decode('utf8'))
        >>> print repr(e)
        ConvertError: u'\u0430\u0431\u0432\u0433\u0434'

        """
        return 'ConvertError: %r' % self.title


class EmptyArticleError(ConvertError): pass

def mkredirect(title, redirect_target):
    meta = {u'r': redirect_target}
    return title, tojson(('', [], meta)), True, None

def convert(title):
    gc.collect()
    try:
        text = wikidb.reader[title]

        if not text:
            raise EmptyArticleError(title)

        redirect = wikidb.get_redirect(text)
        if redirect:
            return mkredirect(title, redirect)

        mwobject = uparser.parseString(title=title,
                                       raw=text,
                                       wikidb=wikidb,
                                       lang=wikidb.lang,
                                       magicwords=wikidb.siteinfo['magicwords'])
        xhtmlwriter.preprocess(mwobject)
        text, tags, languagelinks = writer.convert(mwobject, rtl=wikidb.rtl)
    except EmptyArticleError:
        raise
    except Exception:
        log.exception('Failed to process article %s', title.encode('utf8'))
        raise ConvertError(title)
    else:
        return title, tojson((text.rstrip(), tags)), False, languagelinks


class BadRedirect(ConvertError): pass


def parse_redirect(text, aliases):
    """
    >>> aliases = [u"#PATRZ", u"#PRZEKIERUJ", u"#TAM", u"#REDIRECT"]
    >>> parse_redirect(u'#PATRZ [[Zmora]]', aliases)
    u'Zmora'
    >>> parse_redirect(u'#PATRZ[[Zmora]]', aliases)
    u'Zmora'
    >>> parse_redirect(u'#PRZEKIERUJ [[Uwierzytelnianie]]', aliases)
    u'Uwierzytelnianie'

    >>> parse_redirect('#TAM[[Żuraw samochodowy]]'.decode('utf8'), aliases)
    u'\u017buraw samochodowy'

    >>> parse_redirect('#перенапр[[абв]]'.decode('utf8'), [u"#\u043f\u0435\u0440\u0435\u043d\u0430\u043f\u0440"])
    u'\u0430\u0431\u0432'

    >>> parse_redirect('#Перенапр[[абв]]'.decode('utf8'),
    ...                [u"#\u043f\u0435\u0440\u0435\u043d\u0430\u043f\u0440",
    ...                 u"#\u043f\u0435\u0440\u0435\u043d\u0430\u043f\u0440".upper()])
    u'\u0430\u0431\u0432'

    >>> parse_redirect(u'abc', aliases)

    >>> parse_redirect(u'#REDIRECT [[abc', aliases)
    Traceback (most recent call last):
    ...
    BadRedirect: ConvertError: [[abc

    >>> parse_redirect(u'#REDIRECT abc', aliases)
    Traceback (most recent call last):
    ...
    BadRedirect: ConvertError: abc

    >>> parse_redirect('#REDIRECT абв'.decode('utf8'), aliases)
    Traceback (most recent call last):
    ...
    BadRedirect: ConvertError: абв

    """
    for alias in aliases:
        if text.startswith(alias) or text.upper().startswith(alias):
            text = text[len(alias):].lstrip()
            begin = text.find('[[')
            if begin < 0:
                raise BadRedirect(text)
            end = text.find(']]')
            if end < 0:
                raise BadRedirect(text)
            return text[begin+2:end]
    return None

class Wiki(WikiDB):

    def __init__(self, cdbdir, lang, rtl=False):
        WikiDB.__init__(self, cdbdir, lang=lang)
        self.lang = lang
        self.rtl = rtl
        self.redirect_aliases = set()
        aliases = [magicword['aliases']
                                 for magicword in self.siteinfo['magicwords']
                                 if magicword['name'] == 'redirect'][0]

        for alias in aliases:
            self.redirect_aliases.add(alias)
            self.redirect_aliases.add(alias.lower())
            self.redirect_aliases.add(alias.upper())

    def get_redirect(self, text):
        redirect = parse_redirect(text, self.redirect_aliases)
        if redirect:
            redirect = self.nshandler.get_fqname(redirect)
        return redirect

    def getURL(self, title):
        return ''

    def getSource(self, title, revision=None):
        from mwlib.metabook import make_source

        g = self.siteinfo['general']
        return make_source(
            name='%s (%s)' % (g['sitename'], g['lang']),
            url=g['base'],
            language=g['lang'],
            base_url=self.nfo['base_url'],
            script_extension=self.nfo['script_extension'],
        )

    def normalize_and_get_page(self, name, defaultns):
        fqname = self.nshandler.get_fqname(name, defaultns=defaultns)
        return self.get_page(fqname)

    def normalize_and_get_image_path(self, name):
        assert isinstance(name, basestring)
        name = unicode(name)
        
        ns, partial, fqname = self.nshandler.splitname(name, defaultns=6)
        if ns != 6:
            return

        if "/" in fqname:
            return None


def total(inputfile, options):
    load_siteinfo(options.siteinfo)
    w = Wiki(inputfile, options.wiki_lang)
    for i, a in enumerate(islice(w.articles(), options.start, options.end)):
        pass
    try:
        return i+1
    except:
        return 0


def make_input(input_file_name):
    return input_file_name

def collect_articles(input_file, options, compiler):
    p = WikiParser(options, compiler)
    p.parse(input_file)

siteinfo_loaded = False

def load_siteinfo(filename):
    if siteinfo_loaded:
        return mwlib.siteinfo.get_siteinfo(None)
    if not filename:
        raise Exception('Site info not specified (fetch with aard-siteinfo, '
                        'specify with use --siteinfo)')

    if not os.path.exists(filename):
        raise Exception('File %s not found' % filename)

    with open(filename) as f:
        siteinfo = json.load(f)
        global siteinfo_loaded
        siteinfo_loaded = True

    mwlib.siteinfo.get_siteinfo = lambda lang: siteinfo

    return siteinfo

default_description = """ %(title)s for Aard Dictionary is a collection of text documents from %(server)s (articles only). Some documents or portions of documents may have been omited or could not be converted to Aard Dictionary format. All documents can be found online at %(server)s under the same title as displayed in Aard Dictionary.
"""

class WikiParser():

    def __init__(self, options, consumer):
        self.consumer = consumer
        wiki_lang = options.wiki_lang
        siteinfo = load_siteinfo(options.siteinfo)

        consumer.add_metadata('siteinfo', siteinfo)
        general_siteinfo = siteinfo['general']
        sitename = general_siteinfo['sitename']
        sitelang = general_siteinfo['lang']

        from ConfigParser import ConfigParser
        c = ConfigParser()

        if options.metadata:
            read_metadata_files = c.read(options.metadata)
            if not read_metadata_files:
                log.warn('Metadata file could not be read %s' % options.metadata)
            else:
                log.info('Using metadata from %s', ', '.join(read_metadata_files))
                for opt in c.options('metadata'):
                    value = c.get('metadata', opt)
                    self.consumer.add_metadata(opt, value)
        else:
            log.warn('No metadata file specified')


        if options.license:
            license_file = options.license
        else:
            rights = general_siteinfo['rights']
            if rights in known_licenses:
                license_file = known_licenses[rights]
            else:
                license_file = None
                self.consumer.add_metadata('license', rights)

        if license_file:
            with open(license_file) as f:
                log.info('Using license text from %s', license_file)            
                license_text = f.read()
                self.consumer.add_metadata('license', license_text)            
                
        if options.copyright:
            copyright_file = options.copyright
            with open(copyright_file) as f:
                log.info('Using copyright text from %s', copyright_file)            
                copyright_text = f.read()
                self.consumer.add_metadata('copyright', copyright_text)                        

        self.consumer.add_metadata("title", sitename)
        if options.dict_ver:
            self.consumer.add_metadata("version", 
                                       "-".join((options.dict_ver, 
                                                 options.dict_update)))
        server = general_siteinfo['server']
        self.consumer.add_metadata("source", server)
        self.consumer.add_metadata("description", default_description % dict(server=server, 
                                                                             title=sitename))

        self.lang = wiki_lang
        self.rtl = options.rtl
        self.consumer.add_metadata("lang", wiki_lang)
        self.consumer.add_metadata("sitelang", sitelang)
        self.consumer.add_metadata("index_language", sitelang)
        self.consumer.add_metadata("article_language", sitelang)
        log.info('Language: %s (%s)', self.lang, sitelang)

        self.consumer.add_metadata('mwlib',
                                   '.'.join(str(v) for v in mwlib_version))
        self.processes = options.processes if options.processes else None
        self.pool = None
        self.timeout = options.timeout
        self.timedout_count = 0
        self.start = options.start
        self.end = options.end
        if options.nomp:
            log.info('Disabling multiprocessing')
            self.parse = self.parse_simple
        else:
            self.parse = self.parse_mp
        self.mp_chunk_size = options.mp_chunk_size

        if options.lang_links:
            self.lang_links_langs = frozenset(l.strip().lower()
                                              for l in options.lang_links.split(',')
                                              if l.strip().lower() != sitelang)
            self.consumer.add_metadata("language_links", list(self.lang_links_langs))
        else:
            self.lang_links_langs = frozenset()

        self.requested_article_count = options.article_count


    def articles(self, f):
        if self.start > 0:
            log.info('Skipping to article %d', self.start)
        _create_wikidb(f, self.lang, self.rtl)
        for title in islice(wikidb.articles(), self.start, self.end):
            log.debug('Yielding "%s" for processing', title.encode('utf8'))
            yield title
            gc.collect()

    def reset_pool(self, cdbdir, terminate=True):
        if self.pool and terminate:
            log.info('Terminating current worker pool')
            self.pool.terminate()
        log.info('Creating new worker pool with wiki cdb at %s', cdbdir)

        self.pool = Pool(processes=self.processes,
                         initializer=_init_process,
                         initargs=[cdbdir, self.lang, self.rtl])

    def parse_simple(self, f):
        _init_process(f, self.lang, self.rtl)
        self.consumer.add_metadata('article_format', 'html')
        articles = self.articles(f)
        for a in articles:
            try:
                result = convert(a)
                title, serialized, redirect, langugagelinks = result
                self.consumer.add_article(title, serialized, redirect)
                self.process_languagelinks(title, langugagelinks)
            except EmptyArticleError, e:
                self.consumer.empty_article(e.title)
            except ConvertError, e:
                self.consumer.fail_article(e.title)

    def parse_mp(self, f):
        try:
            self.consumer.add_metadata('article_format', 'html')
            articles = self.articles(f)
            self.reset_pool(f)
            iter_count = 1
            real_article_count = 0
            while True:
                if iter_count:
                    chunk = islice(articles, self.mp_chunk_size)
                    iter_count = 0
                else:
                    break

                resulti = self.pool.imap_unordered(convert, chunk)
                while True:
                    try:
                        result = resulti.next(self.timeout)
                        iter_count += 1
                        title, serialized, redirect, langugagelinks  = result

                        if self.requested_article_count:
                            if  not redirect:
                                real_article_count += 1
                                self.consumer.add_article(title, serialized, redirect)
                                self.process_languagelinks(title, langugagelinks)
                                if real_article_count >= self.requested_article_count:
                                    try:
                                        self.pool.terminate()
                                    except:
                                        log.exception()
                                    finally:
                                        return
                        else:
                            self.consumer.add_article(title, serialized, redirect)
                            self.process_languagelinks(title, langugagelinks)
                    except StopIteration:
                        break
                    except TimeoutError:
                        log.warn('Worker pool timed out')
                        self.consumer.timedout(count=len(multiprocessing.active_children()))
                        self.reset_pool(f)
                        resulti = self.pool.imap_unordered(convert, chunk)
                    except AssertionError:
                        log.exception()
                    except EmptyArticleError, e:
                        self.consumer.empty_article(e.title)
                    except ConvertError, e:
                        self.consumer.fail_article(e.title)
                    except KeyboardInterrupt:
                        log.error('Keyboard interrupt: '
                                  'terminating worker pool')
                        self.pool.terminate()
                        raise

                self.pool.close()
                self.reset_pool(f, terminate=False)
        finally:
            self.pool.close()
            self.pool.join()

    def process_languagelinks(self, title, languagelinks):
        if not languagelinks:
            return
        targets = set()
        for namespace, target in languagelinks:
            if namespace in self.lang_links_langs:
                log.debug('Language link for %s: %s (%s)',
                          title.encode('utf8'), target.encode('utf8'),
                          namespace.encode('utf8'))
                i = target.find(namespace+u':')
                if i > -1:
                    unqualified_target = target[len(namespace)+1:]
                    try:
                        wikidb.reader[unqualified_target]
                    except KeyError:
                        targets.add(unqualified_target)
                else:
                    log.warn('Invalid language link "%s"', target.encode('utf8'))
        for target in targets:
            (l_title, l_serialized,
             l_redirect, l_langugagelinks) = mkredirect(wikidb.nshandler.get_fqname(target), title)
            self.consumer.add_article(l_title, l_serialized,
                                      redirect=True, count=False)




import sys
from setuptools import setup, find_packages

import aardtools

install_requires = ['aarddict >= 0.9.0',
                    'PyICU >= 0.8.1',
                    'mwlib == 0.12.13']

if sys.version_info < (2, 6):
    install_requires += ['simplejson', 'multiprocessing']

setup(
    name = aardtools.__name__,
    version = aardtools.__version__,
    packages = find_packages(),
    entry_points = {
        'console_scripts': ['aardcompile = aardtools.compiler:main',
                            'aardc = aardtools.compiler:main',
                            'aard-siteinfo = aardtools.fetchsiteinfo:main',
                            ]
    },
    install_requires = install_requires,
    package_data={'aardtools': ['licenses/*.txt']},    
    author = "Igor Tkach",
    author_email = "itkach@aarddict.org",
    description =  '''Tools to create dictionaries in aarddict format.''',
    license = "GPL 3",
    keywords = ['aarddict', 'aardtools', 'wiki', 'wikipedia',
                'xdxf', 'dict', 'dictionary', 'maemo'],
    url = "http://aarddict.org",
    classifiers=[
                 'Development Status :: 3 - Alpha',
                 'Operating System :: OS Independent',
                 'Programming Language :: Python',
                 'License :: OSI Approved :: GNU General Public License (GPL)',
                 'Topic :: Utilities',
                 'Environment :: Console'
    ]
)




# This file is part of Aard Dictionary Tools <http://aarddict.org>.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License <http://www.gnu.org/licenses/gpl-3.0.txt>
# for more details.
#
# Copyright (C) 2008-2009  Igor Tkach

import os
import sys
import logging
import functools
from copy import deepcopy

try:
    from xml.etree import cElementTree as etree
except ImportError:
    logging.warning('cElementTree is not available, will use ElementTree')
    from xml.etree import ElementTree as etree

try:
    import json
except ImportError:
    import simplejson as json

try:
    from itertools import combinations
except ImportError:
    def combinations(iterable, r):
        # combinations('ABCD', 2) --> AB AC AD BC BD CD
        # combinations(range(4), 3) --> 012 013 023 123
        pool = tuple(iterable)
        n = len(pool)
        if r > n:
            return
        indices = range(r)
        yield tuple(pool[i] for i in indices)
        while True:
            for i in reversed(range(r)):
                if indices[i] != i + n - r:
                    break
            else:
                return
            indices[i] += 1
            for j in range(i+1, r):
                indices[j] = indices[j-1] + 1
            yield tuple(pool[i] for i in indices)
    globals()['combinations'] = combinations


tojson = functools.partial(json.dumps, ensure_ascii=False)


def total(inputfile, options):
    count = 0
    for event, element in etree.iterparse(inputfile):
        if element.tag == 'ar':
            keys = element.findall('k')
            for key_element in keys:
                n_opts = len([c for c in key_element if c.tag == 'opt'])
                if n_opts:
                    for j in range(n_opts+1):
                        for comb in combinations(range(n_opts), j):
                            count += 1
                else:
                    count += 1
        if element.tag != 'k':
            element.clear()
    return count

def make_input(input_file_name):
    if input_file_name == '-':
        return sys.stdin
    import tarfile
    try:
        tf = tarfile.open(input_file_name)
    except:
        #probably this is not tar archive, open regular file
        return open(input_file_name)
    else:
        for tar in tf:
            if os.path.basename(tar.name) == 'dict.xdxf':
                return tf.extractfile(tar)
    raise IOError("%s doesn't look like a XDXF dictionary" % input_file_name)

def collect_articles(input_file, options, compiler):
    p = XDXFParser(compiler, options)
    p.parse(input_file)


xdxf_visual_tags = frozenset(('ar',
                              'k',
                              'opt',
                              'nu',
                              'def',
                              'pos',
                              'tense',
                              'tr',
                              'dtrn',
                              'kref',
                              'rref',
                              'iref',
                              'abr',
                              'c',
                              'ex',
                              'co',
                              'su'))


class XDXFParser():

    def _tag_handler_ar(self, e, **_):
        e.set('class', e.tag)
        e.tag = 'div'

    def _tag_handler_c(self, child, **_):
        child.tag = 'span'
        color = child.get('c', '')
        child.attrib.clear()
        child.set('style', 'color: %s;' % color)

    def _tag_handler_iref(self, child, **_):
        child.tag = 'a'

    def _tag_handler_kref(self, child, **_):
        child.tag = 'a'
        child.set('href', child.text)

    def _tag_handler_su(self, child, **_):
        child.tag = 'div'
        child.set('class', 'su')

    def _tag_handler_def(self, child, **_):
        child.tag = 'blockquote'

    def _tag_handler_abr(self, child, **kw):
        abbreviations = kw['abbreviations']
        child.tag = 'abbr'
        abr = child.text
        if abr in abbreviations:
            child.set('title', abbreviations[abr])

    def default_tag_handler(self, child, **_):
        if child.tag in xdxf_visual_tags:
            child.set('class', child.tag)
            child.tag = 'span'

    def __init__(self, consumer, options):
        self.consumer = consumer
        self.options = options

    def _mkabbrs(self, element):
        abbrs = {}
        for abrdef in element:
            if abrdef.tag.lower() == 'abr_def':
                value = abrdef.find('v')
                if value:
                    value_txt = value.text
                    for key in abrdef.findall('k'):
                        abbrs[key.text] = value_txt
        return abbrs

    def _transform_element(self, element, abbreviations):
        handler = getattr(self, '_tag_handler_'+element.tag.lower(), self.default_tag_handler)
        handler(element, abbreviations=abbreviations)


    def _text(self, xdxf_element, abbreviations):
        element = deepcopy(xdxf_element)
        if self.options.skip_article_title:
            tail = ''
            for k in list(element.findall('k')):
                if k.tail:
                    tail += k.tail
                element.remove(k)
            tail = tail.lstrip()
            element.text = tail + element.text if element.text else tail
        self._transform_element(element, abbreviations)
        for child in element.getiterator():
            self._transform_element(child, abbreviations)
        return etree.tostring(element, encoding='utf8')

    def _mktitle(self, title_element, include_opts=()):
        title = title_element.text
        opt_i = -1
        for c in title_element:
            if c.tag == 'nu' and c.tail:
                if title:
                    title += c.tail
                else:
                    title = c.tail
            if c.tag == 'opt':
                opt_i += 1
                if opt_i in include_opts:
                    if title:
                        title += c.text
                    else:
                        title = c.text
                if c.tail:
                    if title:
                        title += c.tail
                    else:
                        title = c.tail
        return title

    def parse(self, f):
        self.consumer.add_metadata('article_format', 'html')
        abbreviations = {}
        for _, element in etree.iterparse(f):
            if element.tag == 'description':
                self.consumer.add_metadata(element.tag, element.text)
                element.clear()

            if element.tag == 'full_name':
                self.consumer.add_metadata('title', element.text)
                element.clear()

            if element.tag == 'xdxf':
                self.consumer.add_metadata('article_language',
                                           element.get('lang_to'))
                self.consumer.add_metadata('index_language',
                                           element.get('lang_from'))
                self.consumer.add_metadata('xdxf_format',
                                           element.get('format'))
                element.clear()

            if element.tag == 'abbreviations':
                abbreviations = self._mkabbrs(element)

            if element.tag == 'ar':
                txt = self._text(element, abbreviations)
                txt = txt.replace('\n', '<br/>')
                titles = []
                for title_element in element.findall('k'):
                    n_opts = len([c for c in title_element if c.tag == 'opt'])
                    if n_opts:
                        for j in range(n_opts+1):
                            for comb in combinations(range(n_opts), j):
                                titles.append(self._mktitle(title_element, comb))
                    else:
                        titles.append(self._mktitle(title_element))

                if titles:
                    first_title = titles[0]
                    serialized = tojson((txt, [], {}))
                    self.consumer.add_article(first_title, serialized,
                                              redirect=False)
                    titles = titles[1:]
                    if titles:
                        for title in titles:
                            logging.debug('Redirect %s ==> %s',
                                          title.encode('utf8'),
                                          first_title.encode('utf8'))
                            meta = {u'r': first_title}
                            serialized = tojson(('', [], meta))
                            self.consumer.add_article(title, serialized,
                                                      redirect=True)
                else:
                    logging.warn('No title found in article:\n%s',
                                 etree.tostring(element, encoding='utf8'))
                element.clear()
